/*
SAS Genetic Marker Analytical Process

Name: Linkage Disequilibrium
Author: Wendy Czika

Description:
Linkage Disequilibrium calculates various LD measures,
as well as an overall test for LD between
pairs of SNPs, using PROC ALLELE.  Two
types of plots are created to represent
different LD measures: a triangular plot
with colors of symbols representing the
p-values for LD between pairs or markers,
and the HWD for individual markers (with
red=pvalue<0.01, orange=0.01<pvalue<0.05, 
and yellow=0.05<pvalue); and a contour plot
representing ranges of either the square
of the correlation coefficient or absolute
value of D'.  See the ALLELE procedure
in the SAS/Genetics User's Guide for more
information.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%let InData = ; 
* Full path and name to the input SAS data set. 
The name must be lowercase.;

%let AlleleVars = ; 
* Variables containing marker alleles (2 per marker).;

%let ByVars = ; 
* BY variables for PROC ALLELE.;

%let WhereStmt = ;
* Optional WHERE statement for PROC ALLELE.;

%let AnnoInData = ;
* Annotation SAS data set containing marker information.;

%let AnnoLabel = ; 
* Annotation variable to be used as label for markers 
in output and plots.;

%let AnnoGroup = ;
* Annotation variable for grouping analyses.;

%let AnnoLoc = ;
* Annotation variable (numeric) containing location of markers .;

%let AnnoAccession = ; 
* Annotation variable containing accession number for web searches.;

%let LDMeasures = CORRCOEFF DELTA DPRIME PROPDIFF YULESQ;
* Any subset of these 5 measures can be specified.;

%let Maxdist = 50;
* Maximum number of neighboring markers to compute LD with each marker.;

%let Haplo = EST;
* Method of haplotype estimation: NONE, EST, or GIVEN.
Default is EST which uses the maximum likelihood estimate
of the haplotype frequency.;

%let ChartHaploFreqs = No;
* Specify 'Yes' to create frequency charts of all two-locus
haplotype frequencies;

%let ContPlotMeas = Dprime;
* LD Measure to be plotted: Corrcoeff
for the square of the correlation coefficient,
or Dprime for the absolute value of D'.
Only plotted if all markers are biallelic.;

%let NumCont = 9;
* Number of contours to use in contour plot.;

%let OutPath = ; 
* Path for output on your SAS server. If you leave this blank the system 
will use a default location.;

%let MacroPath = ; 
* Path for SAS macros.;

%let KeepProcessInfo = No;
* Option to keep process information.  If specify "Yes", a data set, 
_processInfo, will be add to result tree;
 
*ProcessBody;

%macro LinkageDisequilibrium;

*initialization and argument checking;
%let ProcessName = ld; 
%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool nolabel;
data _null_;
   infile "&MacroPath.UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for SAS macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;

%PathName(&InData);
%let InData = &TmpName;
libname InLib "&TmpPath";

%CheckForBlank(&OutPath,Output path);
libname OutLib "&OutPath";

%let AnnoMergeVar=&AnnoLabel;
%CheckAnno(checkindata=0);

%if %length(&ByVars) %then %do;
 proc sort data=InLib.&InData out=a;
   by &ByVars;
   %if %length(&WhereStmt) %then %do;
     where &WhereStmt;
   %end;
 run;
 %let ds = a;
 %let ByVars=%upcase(&ByVars);
 %let temp=%sysfunc(tranwrd(&ByVars,DESCENDING, ));
 %let ByVarsOnly=%sysfunc(compbl(&temp)); 
 %let JBy=%sysfunc(translate(&ByVarsOnly,","," "));
 %let i=1;
 %do %while (%length(%scan(&ByVarsOnly, &i))>0);
  %let i=%eval(&i+1);
 %end;
 data countby;
  set &ds;
  by &ByVars;
  %let lastby=%scan(&ByVarsOnly, %eval(&i-1));
  if first.&lastby;
 run;
 %let dsid=%sysfunc(open(countby));
 %let Nbyval=%sysfunc(attrn(&dsid,NOBS));
 %let ret=%sysfunc(close(&dsid));	
%end;

%else %do;
 %let ds = InLib.&InData;
 %let Nbyval=1;
 %let ByVarsOnly = ;
%end;

options label;

data count; 
 set &ds;
 keep &AlleleVars;
run;

%let dsid=%sysfunc(open(count));
%let n2snps=%sysfunc(attrn(&dsid,NVARS));
%let ret=%sysfunc(close(&dsid));
%let nummark=%eval(&n2snps/2);

%let LDMeasures=&LDMeasures &ContPlotMeas;

%if %length(&AnnoGroup)=0 %then %let NumGps=1;
%else %let NumGps = ;

%if &NumGps=1 %then %do;
 %let AnnoGroup=_group_;
 %if %length(&AnnoInData) %then %do;
  data mapdata;
   set AnnoLib.&AnnoData;
   _group_=1;
   name=&AnnoLabel;
   Marker1=name;
  run;
 %end;
 %else %do;
  %let AnnoLoc=MarkerNumber;
  data mapdata;
   do i=1 to &nummark;
    name="Marker"||left(i);
	Marker1=name;
	MarkerNumber=i;
	_group_=1;
    output;
   end;
  run;
 %end;
%end;
%else %do;
 data mapdata;
  set AnnoLib.&AnnoData;
  name=&AnnoLabel;
  Marker1=name;
 run;
%end;

data _null_;
  set mapdata;
  call symput('HapGp'||left(_n_),trim(left(&AnnoGroup)));
run;
	
data ldmeasures;
 delete;
run;
		
data ldstats;
 delete;
run;

proc sort data=mapdata out=mapgp nodupkey;
 by &AnnoGroup;
run;

data _null_;
 set mapgp;
 call symput('Groupn'||left(_n_),trim(left(&AnnoGroup)));
run;

%let dsid=%sysfunc(open(mapgp));
%let NumGps=%sysfunc(attrn(&dsid,NOBS));
%let ret=%sysfunc(close(&dsid));
 
%let dsid=%sysfunc(open(mapdata));
%let GpPos=%sysfunc(varnum(&dsid,&AnnoGroup));
%let GpType=%sysfunc(vartype(&dsid,&GpPos)); 
%let rc1=%sysfunc(close(&dsid));  

%do i=1 %to &NumGps;
 %let HGVars&i = ;
 %do j=1 %to &nummark;
  %if &&HapGp&j=&&Groupn&i %then %do;
   %let j1=%eval(2*&j-1);
   %let j2=%eval(2*&j);
   %let HGVars&i = &&HGVars&i _a_&j1 _a_&j2;
  %end;
 %end;
		
 data map_sub;
  set mapdata;
  %if (&GpType=C) %then %do;
  	 where &AnnoGroup="&&Groupn&i";
  %end;
  %else %do;
     where &AnnoGroup=&&Groupn&i;
  %end;
 run;		 
	
 data indat;
  set &ds;
  array av{&n2snps} &AlleleVars;
  array _a_{&n2snps} $;
  do i=1 to &n2snps;
   _a_{i}=compress(av{i});
  end;
  drop &AlleleVars;
 run;

 %let snp&i=0;
 %if %length(&&HGVars&i) %then %do;
  * setup for html output;
  ods output LDMeasures=ldm MarkerSumm=ms;

  proc allele data=indat ndata=map_sub nofreq &LDMeasures outstat=out
            haplo=&Haplo maxdist=&Maxdist; 
   %if %length(&ByVars) %then %do; 
    by &ByVars;
   %end;
   var &&HGVars&i;  
   %if %length(&WhereStmt) %then %do;
    where &WhereStmt;
   %end;
  run;

  %if (&syserr>4) %then %goto exit;

  proc summary data=ms;
   var NAllele;
   output out=checkbi max=amax;
  run;
  
  data _null_;
   set checkbi;
   call symput ('amax',left(amax));
  run; 

  %if (&amax=2) %then %let snp&i=1;

  data ldmeasures;
   set ldmeasures ldm(in=new);
   %if (&GpType=C) %then %do;
     if new then &AnnoGroup="&&Groupn&i";
   %end;
   %else %do;
     if new then &AnnoGroup=&&Groupn&i;
   %end;
 *  if new=0 or (locus2 ne lag(locus2) or locus1 ne lag(locus1));
  run;
		
  data ldstats;
   set ldstats out(in=new);
    %if (&GpType=C) %then %do;
     if new then &AnnoGroup="&&Groupn&i";
   %end;
   %else %do;
     if new then &AnnoGroup=&&Groupn&i;
   %end;
  run;
	
 %end;
%end;
	
ods html 
   body = "&ProcessName.body.html"
   frame = "&ProcessName.frame.html"   
   contents = "&ProcessName.contents.html"
   path = "&OutPath" (url=none)
;

proc print data=ldmeasures label;
 title 'LD Measures';
run;

proc print data=ldstats label;
 title 'Test for LD';
run;

ods html close; 
 
options nolabel;

proc sort data=ldstats;
 by Locus1;
run;

%if %upcase(&ChartHaploFreqs)=YES %then %do;
 data ldm;
  set ldmeasures;
  sortorder=_n_;
 run;

 proc sort data=ldm;
  by &ByVars Locus1 Locus2;
 run;

 proc sort data=ldstats out=ldtot;
  by &ByVars Locus1 Locus2;
 run;

 data ld2;
  merge ldm(in=k) ldtot;
  by &ByVars Locus1 Locus2;
  Count=round(2*NIndiv*Freq);
  drop NIndiv ChiSq DF Test ProbChi;
  if k;
 run;
 
 proc sort data=ld2 out=ldmeasures(drop=sortorder);
  by sortorder;
 run;
%end;

proc sort data=ldmeasures out=OutLib.ldmeasures(rename=(Locus1-Locus2=Marker1-Marker2));
 by Locus1;
run;

data ldstats;
 set ldstats;
 rename Locus1-Locus2=Marker1-Marker2;
run;

%if %length(&AnnoLoc)=0 %then %do;
 proc sort data=mapdata;
	 by &AnnoGroup;
	run;
	
	data mapdata;
	 set mapdata;
	 by &AnnoGroup;
	 if first.&AnnoGroup then MarkerNumber=1;
	 else MarkerNumber+1;
	run;
 %let AnnoLoc=MarkerNumber;
%end;
		
proc sort data=mapdata;
 by Marker1;
run;

data temp1;
 merge ldstats(in=inc) mapdata;
 by Marker1;
 Locus1=&AnnoLoc+0;
 label Locus1 = ' ';
 drop &AnnoLoc;
 if inc;
run;

data temp1m; 
 merge OutLib.ldmeasures(in=inc) mapdata;
 by Marker1;
 Locus1=&AnnoLoc+0;
 label Locus1 = ' ';
 drop &AnnoLoc;
 l1=lag(Marker1);
 l2=lag(Marker2);
 if Marker1 ne l1 or Marker2 ne l2;
 if inc;
run;

proc sort data=temp1;
 by Marker2;
run;

proc sort data=temp1m;
 by Marker2;
run;

data mapdata;
 set mapdata;
 Locus2=&AnnoLoc+0;
 keep Locus2 Marker1;
 rename Marker1=Marker2;
run;

data temp2;
 merge temp1(in=inc) mapdata;
 by Marker2;
 label ProbChi=' ';
 if inc;
 %if %length(&AnnoAccession)>0 %then %do;
  if Test ne "HWE" then &AnnoAccession=' ';
 %end;
 keep &ByVarsOnly Locus1 Locus2 Marker1 Marker2 Test ProbChi &AnnoGroup &AnnoAccession;
run;

data temp2m;
 merge temp1m(in=inc) mapdata;
 by Marker2;
 if inc;
 keep &ByVarsOnly Locus1 Locus2 Marker1 Marker2 &ContPlotMeas &AnnoGroup;
run;

%if %length(&ByVars) %then %do;
 proc sort data=temp2;
  by &ByVars;
 run;
			
 proc sort data=temp2m;
  by &ByVars;
 run;
%end;

data OutLib.ldstats;
 set temp2;
 temp=Locus1;
 tl=Marker1;
 if Locus1 ge Locus2 then do;
  Locus1=Locus2+0;
  Locus2=temp;
  Marker2=Marker1;
  Marker1=tl;
 end;
 drop temp tl;
run;

%let ContPlotMeas=%upcase(&ContPlotMeas);

data OutLib.contplot;
 set temp2m temp2m(in=switch);
 temp=Locus1;
 tl=Marker1;
 if switch then do;
  Locus1=Locus2+0;
  Locus2=temp;
  Marker1=Marker2;
  Marker2=tl;
 end;
 %if (&ContPlotMeas=DPRIME) %then %do;
  dprime=abs(dprime);
 %end;
 %else %do;
  corrcoeff=corrcoeff*corrcoeff;
 %end;
 %let i=1;
 label &ContPlotMeas=' ';
 drop temp tl;
run;
 
%let lowpcut = 0.01;
%let midpcut = 0.05;

* create JSL file;
%let JSLFileName = &ProcessName..jsl;
filename jslfile "&OutPath.&JSLFileName";

* put JSL statements;
 data _null_;
   file jslfile;
   put "nmarkers = &nummark;";
   put "fs=15;";
   put "%str(data=open(%".\ldstats.sas7bdat%");)";
   %if %length(&AnnoInData) %then %do;
      %include "&MacroPath.WebSearchJSL.sas";
      %WebSearchJSL;
   %end;
   put "Pref(Outline Markers by Default(1));";
   put 'column("Locus2") << label(1);';
   put 'fs=5;';
   put 'column("Marker2") << label(1);';
   put 'column("Locus1") << label(1);';
   put 'column("Marker1") << label(1);';
   put 'column("Test") << label(1);';
   put 'column("ProbChi") << label(1);';
   %do i=1 %to &NumGps;
 	%if %length(&&HGVars&i) %then %do;
     put 'tri = New Window("LD Triangle Plot' @;
     %if &NumGps>1 %then %do;
      put "%str( for Group &&Groupn&i%",)";
	 %end;
	 %else %do; 
      put '",';
	 %end;
     put 'op = Overlay Plot(X( :Locus2), Y( :Locus1)';
     %if &NumGps>1 %then %do;
      %if &GpType=C %then %do;
       put "%str(, Where%(:&AnnoGroup==%"&&Groupn&i%"%))";
      %end;
      %else %do;
       put ",Where(:&AnnoGroup==&&Groupn&i)";
      %end;
	 %end;
     %if %length(&ByVars)>0 %then %do;
	  put ", By(&JBy)	));";
	 %end;
	 %else %do;
	  put ' )); ';
	 %end;
	 %if %length(&ByVars)=0 %then %do;
	  put 'tri[outlinebox(1)]<<SetTitle("Locus1 versus Locus2");';
	 %end;
	 %do j=1 %to &Nbyval;
	  put "tri[framebox(&j)] << FrameSize(fs*nmarkers,fs*nmarkers);";
      put "tri[framebox(&j)] << MarkerSize(8);";
      put "tri[framebox(&j)] << ScrollWindow(0,20*nmarkers);";
	 %end;
     put 'for each row(ColorOf(RowState())=9);';
     put "for each row(if((ProbChi <= &lowpcut), ColorOf(RowState())=3));";
     put "for each row(if(((ProbChi > &lowpcut) & (ProbChi <= &midpcut)),ColorOf(RowState())=6));";
     put "for each row(if((ProbChi > &midpcut), ColorOf(RowState())=9));";
	%end;	
   %end;
   %if %length(&ContPlotMeas)>0 %then %do;
    put 'data=open(".\ContPlot.sas7bdat");';
	%do i=1 %to &NumGps;
	 %if (%length(%scan(&&HGVars&i,5)) and &&snp&i=1) %then %do;
	  put "%str(tri = New Window%( %"Contour Plot of &ContPlotMeas )" @;
      %if &NumGps>1 %then %do;
       put "%str( for Group &&Groupn&i%",)";
	  %end;
	  %else %do; 
       put "%str(%",)";
	  %end;
      put "Contour Plot(X( :Locus1, :Locus2), Y( :&ContPlotMeas),";
      put "Show Data Points(0), Fill Areas(1), Label Contours(0),";
      put " Specify Contours(Min(0), Max(1), N(&NumCont))";
	  %if &NumGps>1 %then %do;
       %if &GpType=C %then %do;
        put "%str(, Where%(:&AnnoGroup==%"&&Groupn&i%"%))";
       %end;
       %else %do;
        put ",Where(:&AnnoGroup==&&Groupn&i)";
       %end;
	  %end;
	  %if %length(&ByVars)>0 %then %do;
	   put ", By(&JBy))); ";
	  %end;
	  %else %do;
	   put ' )); ';
	  %end;
	 %end;
	%end;
   %end;
   %if %upcase(&ChartHaploFreqs)=YES %then %do;
    put 'data=open(".\ldmeasures.sas7bdat");'; 
    put 'cp=New Window( "Haplotype Frequencies",';
    put 'Chart(Freq( :Count), X(Haplotype), Y(N), Show Y Legend(0),';
    put 'Show Level Legend(1), Bar Chart(1)';
    put ', By(Marker1, Marker2)';
    %if %length(&ByVars)>0 %then %do;
 	 put ", Grouping(&JBy))); ";
    %end;
    %else %do; 
     put ')); ';
    %end;
   %end;
run;

* Packaging;
%PackageBegin(&ProcessName);
%PackageInsertHtml(html from &ProcessName, &OutPath, &ProcessName.body.html,
   &ProcessName.frame.html, &ProcessName.contents.html);
%if %upcase(&KeepProcessInfo) = YES %then %do;
   %PackageInsertDataset(process information of &ProcessName, outlib, _processinfo);
%end;
%PackageInsertDataset(data from &ProcessName, OutLib, ldstats);
%PackageInsertDataset(data from &ProcessName, OutLib, ldmeasures);
%PackageInsertDataset(contour plot data, OutLib, ContPlot);
%if %length(&ContPlotMeas)>0 %then %do;
 %PackageInsertFile(JSL Code from &ProcessName, &OutPath, &JSLFileName, "ldstats.sas7bdat,contplot.sas7bdat,ldmeasures.sas7bdat", text/plain, yes);
%end;
%else %do;
	%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &JSLFileName, "ldstats.sas7bdat,ldmeasures.sas7bdat", text/plain, yes);
%end;
%PackagePublish(&OutPath,to_requester);
%PackageEnd;

%exit:;
%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %do;
   %ClockStop(&ProcessName)
%end;

%mend;

* run the macro;
%LinkageDisequilibrium;
































































































































































