/*
SAS Genetic Marker Analytical Process

Name: NEXUS Input Engine
Author: Xiang Yu, Russ Wolfinger, SAS Institute Inc.
Support: Xiang Yu, Russ Wolfinger, SAS Institute Inc.
History:
   sasrdw 08Dec03 initial coding                          

Description: 
The NEXUS Input Engine imports genetic
marker data in the NEXUS format into a SAS data set and
uploads it to the warehouse.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%let InFile = ; * Name of input NEXUS file with full path;

%let OutData = ; 
* Name of SAS data set created from the NEXUS file and uploaded to the warehouse.;

%let MarkerData = ; 
* Name of SAS data containing marker names.  This can be null if there is
no description about markers in the input file;

%let OutPath = ; 
* Path for output on your SAS server. If you leave this blank the system 
will use a default location.;

%let UploadRawFile = No;
* Specify "Yes" to upload InFile (in original format) into warehouse;

%let MacroPath = ; 
* Path for SAS macros.;

*ProcessBody;

*Parse Nexus format files;
%macro ParseNexus;
*initialization;
%let ProcessName = NexusInputEngine;

%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool;
%if &syserr %then %do;
   %put ERROR: Path for utility macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;
/*
%CheckForBlank(&DataPath, Full path of data);
%if &exiterror %then %goto exit;
%CheckForBlank(&OutPath, Output path);
%if &exiterror %then %goto exit; */

%CheckForBlank(&InFile, nexus file);
%if &exiterror %then %goto exit;

%PathName(&InFile);
%if &exiterror %then %do;
   %put ERROR: Nexus file specification is invalid.;
   %goto exit;
%end;

* create output libaray;
libname OutLib "&OutPath";

%CheckDataName(_name=OutData,_default=&ProcessName);

%if &sysscp=WIN %then %let _slash=\;
%else %let _slash=%str(/);

%include "&MacroPath.UtilityMacros.sas" / nosource;    
%let VeryStart = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%let Day1 = %sysfunc(round(%sysevalf(%sysfunc(date()))));

*Parsing input nexus file;
/* %let flagdot = 0;
%if (%length(&InFile)>4) & (%upcase(%substr(&InFile,%length(&InFile)-3,4))=.NEX) %then %do;
%let datafile=&DataPath.&InFile;
%let flagdot = 1;
%end; 
%if &flagdot = 0 %then %do; 
	filename DataFil1 "&DataPath.&InFile..NEX";
	filename DataFil2 "&DataPath.&InFile..nex";
%if %sysfunc(fexist(DataFil1)) %then %let Datafile=&DataPath.&InFile..NEX;
%else %if %sysfunc(fexist(DataFil2)) %then %let Datafile=&DataPath.&InFile..nex;
%else %put ERROR: &DataPath.&InFile dose not exist.;
%end; */
%let Datafile = &Infile;

%let FileStartRow = 0;
%let DataBlockRow = 0;
%let LocusLabelRow = 0;
%let MissingAllele = .;

/* Parse headers */
data _null_;
	InFile "&Datafile" TRUNCOVER EXPANDTABS lrecl=32767 firstobs=1; 
	LENGTH var1 $ 256;
	input var1 1-256;
	var1 = upcase(var1);
	if substr(left(var1),1,6) = '#NEXUS' then call symput('FileStartRow', _n_);
	if scan(var1, 1) = 'BEGIN' & upcase(scan(var1, 2)) = 'GDADATA' then call symput("DataBlockRow", _n_);
	if scan(var1, 1) = 'DIMENSIONS' then 
		do;			
			do i = 2 to 3;
				des = scan(var1, i);
				if index(des, 'NLOCI=') ^= 0 then nloci = substr(des, index(des, 'NLOCI=')+6);
				if index(des, 'NPOPS=') ^= 0 then npops = substr(des, index(des, 'NPOPS=')+6);				
			end;
			call symput('NLoci', nloci);
			call symput('NPop', npops);
		end;
	if scan(var1, 1) = 'FORMAT' then 
		do;			
			if index(var1, 'MISSING=') ^= 0 then missinglabel = substr(var1, index(var1, 'MISSING=')+8, 1);
				else missinglabel = '.';
			if index(var1, 'SEPARATOR=') ^= 0 then separator = substr(var1, index(var1, 'SEPARATOR=')+10, 1);
				else separator = '/';
			if index(var1, 'HAPLOID') ^= 0 then isHaploid = 1;
				else isHaploid = 0;
			if index(var1, 'INTERLEAVE') ^= 0 then isInterleaved = 1;
				else isInterleaved = 0;
			call symput('MissingLabel', trim(left(missinglabel)));
			call symput('Separator', trim(left(separator)));
			call symput('IsHaploid', isHaploid);
			call symput('IsInterleaved', isInterleaved);
		end;
	if scan(var1, 1) = "LOCUSALLELELABELS" then	call symput("LocusLabelRow", _n_);
	if scan(var1, 1) = "MATRIX" then
		do;
			call symput("DataStartRow", _n_);
			stop;
		end;
run;
%if &FileStartRow = 0 %then %put ERROR: Invalid nexus file, must begin with "#nexus";
%if &DataBlockRow = 0 %then %put ERROR: Invalid nexus file, can not find data block starting with command "begin gdadata;";
%if &DataStartRow = 0 %then %put ERROR: Invalid nexus file, can not find data matrix starting with command "matrix";

/* Get marker names */
/* No locus label, use default */
/* %if &LocusLabelRow = 0 %then 
%do i = 1 %to &NLoci;
	Marker&i = "M&i";
%end; */

/* Use locus labels */
%if &LocusLabelRow ^= 0 %then
	%do;
		data markers; 
			InFile "&Datafile" firstobs=%eval(&LocusLabelRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
			length v1 $ 512;
			input v1 1-512;	
			if _n_ = &NLoci then call symput('MarkerNameLen', length(trim(left(v1))));
			if _n_ > &NLoci then stop;
		run;
		data markername; set markers;
			retain namelen 0;
			length mk1 $ &MarkerNameLen mk $ &MarkerNameLen;
            v1=compbl(v1);
            b=indexc(v1, ' ,;:');
            sub=substr(v1, b+1);
			if index(sub,"'")>0 then
             mk1=scan(substr(sub,index(sub,"'")),1,"'");
            else if index(sub,'"')>0 then 
             mk1=scan(substr(sub,index(sub,'"')),1,'"');
            else
			 mk1 = scan(v1, 2, " ,:;");
			*remove spaces to ensure valid variable name;
			mk = translate(trim(mk1), '_',' ');
			namelen = max(namelen, length(trim(left(mk))));
			keep mk;
			if _n_ = &NLoci then call symput('MarkerNameLen', namelen);
		run;
		%if %length(&markerdata) > 0 %then
			%do;
			%CheckDataName(_name=markerdata);

				data OutLib.&MarkerData; set markername;
					length name $ &MarkerNameLen;
					name = mk;
					drop mk;
				run;
			%end;
		/* proc transpose data = markername out=name (drop = _name_);
			var mk;
		run;
		data _null_; set name;
			%do i = 1 %to &NLoci;
				call symput("Marker&i", trim(left(col&i)));
			%end;			 
		run;		 */
	%end;
%else %let MarkerData = ;

*data format check;
data _null_;
	InFile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
	length v1 $ 1024;
	retain hasFreq hasID dataStartAtPopulationLabel 0;			
	input v1 1-1024;
	if index(v1, ';') ^= 0 then stop;
	if &IsHaploid then
		do;
			if index(v1, ':') ^= 0 & dataStartAtPopulationLabel = 0 then
				do;
					vtemp = trim(substr(v1, index(v1, ':')));
					if trim(left(scan(vtemp, &NLoci, " ,"))) ^= '' then dataStartAtPopulationLabel = 1;
				end;
			if index(v1, ':') = 0 then
				do; 
					if trim(left(scan(v1, %eval(&NLoci+1), " ,"))) ^= '' then hasID = 1;
				end;
			else if trim(left(scan(v1, 2, " ,"))) ^= '' then
				do;
					if trim(left(scan(v1, %eval(&NLoci+2), " ,"))) ^= '' then hasFreq = 1;
				end;
			call symput('DataStartAtPopulationLabel', dataStartAtPopulationLabel);
			call symput('HasID', hasID);
			call symput('HasFreq', hasFreq);
			if hasFreq & hasID then stop;
		end;
	else do;
		*hasFreq = 0;
		if index(v1, ':') ^= 0 & dataStartAtPopulationLabel = 0 & index(v1, "&Separator") ^= 0 
			then dataStartAtPopulationLabel = 1;
		if index(v1, "&Separator") ^= 0 then 
			do;			
				if index(v1, ':') ^= 0 then hasFreq = 1;
				var1 = scan(v1, 1, " ,"); 
				if index(var1, "&Separator") ^= 0 then hasID = 0;
					else hasID = 1;
				call symput('DataStartAtPopulationLabel', dataStartAtPopulationLabel);
				call symput('HasID', hasID);
				call symput('HasFreq', hasFreq);
				if hasFreq then stop;
			end;	
		end;	
run;

%let nvar = %eval(&NLoci + &HasID + &HasFreq);
data gdadata; 
	InFile "&Datafile" firstobs=%eval(&DataStartRow+1) TRUNCOVER EXPANDTABS lrecl=32767; 
	length v1 $ 1024;
	input v1 1-1024;
	length charfirst $ 1 charlast $ 1;
	retain pop;
	retain poplen 0 varlen1-varlen&nvar 0 ; 
	
	* Population description;
	if (&IsHaploid = 0 & index(v1, ':') ^= 0 & index(v1, "&Separator") = 0) | (&IsHaploid & index(v1, ':') ^= 0) then 
		do;
			pop = substr(v1, 1, index(v1, ':')-1);
			popLen = max(poplen, length(trim(left(pop))));
			if &DataStartAtPopulationLabel then 
				do;
					flag = 0;
					v1 = trim(left(substr(v1, index(v1, ':'))));
					do;
						%do i = 1 %to &nvar;
							var&i = scan(v1, &i, " ,:");
							varlen&i = max(varlen&i, length(trim(left(var&i))));
						%end;
					end;
				end;
			else
				do;
					flag = 1;
				end;
		end;
	*Marker Data;
	else
		do;
			flag = 0;		
			* delete comments;
			vstripped = trim(left(v1));
			charfirst = substr(vstripped, 1, 1);
			charlast = substr(vstripped, length(vstripped), 1);
			if charfirst = "[" & charlast = "]" then flag = 1;		
			do;
				%do i = 1 %to &nvar;
					var&i = scan(v1, &i, " ,:");
					varlen&i = max(varlen&i, length(trim(left(var&i))));
				%end;
			end;
			if trim(left(var1)) = '' & trim(left(var&nvar)) = '' then flag = 1;
		end;
	*end of data block;
	if index(v1, ';') ^= 0 then
		do;
			call symput('PopLen', poplen);			
			%do i = 1 %to &NLoci;
				%let j = %eval(&i + &HasID);
				call symput("MLen&i", varlen&j);
			%end; 
			if &HasID then call symput('IDLen', varlen1);
			stop;	
		end;
	if &HasFreq then freqstr = var&nvar;
		else freqstr = '';
	if freqstr = '' | index(freqstr, '[') ^= 0 then freqstr = '1';
	drop v1 vstripped charfirst charlast poplen varlen1-varlen&nvar;
run;

*Handle interleaved data;
%if &IsInterleaved %then
	%do;
		%let nvar = %eval(&nvar - &HasFreq);
		data gdadata; set gdadata;
			if flag = 1 then delete;
			drop flag;
		run;
		%NObsVars(gdadata);
		data _null_; set gdadata(keep = pop);
			retain npop 0;
			if pop ^= lag(pop) then npop = npop + 1;
			if _n_ = &NObs then call symput('NPopFound', npop);
		run;
		%let NSection = %sysevalf(&NPopFound/&NPop);
		%let NInds = %sysevalf(&NObs/&NSection);
		%let offset = 0;
		data popfreqstr; set gdadata(keep=pop freqstr);
			if _n_ <= &Ninds;
		run;
		
		%let mergestring = ;
		%do i = 1 %to &NSection;
			%let mergestring = &mergestring section&i ;
			%let start = %sysevalf((&i - 1)*&NInds + 1);
			%let end = %sysevalf(&i*&NInds);			
			data section&i; set gdadata(drop = pop freqstr);
				if _n_ >= &start & _n_ <= &end;				
			run;
			data _null_; set section&i(obs=1);
				array v(&nvar) var1-var&nvar;
				j = 1;
				do while (j <= &nvar & trim(left(v(j))) ^= '');
					j = j+1;
 				end;	
				call symput('varend', trim(left(j)));	
			run; 
			%if &HasID & &i > 1 %then
				%do;
					%let end = %eval(&varend -1);
					data section&i; set section&i;
						drop var1 var&varend-var&nvar;
						%do j = 2 %to &end;
							%let k = %eval(&offset + &j);
							rename var&j = var&k;
							%let MLen&k = &&MLen&j; 
						%end;
					run;	
				%end;	
			%else %if &i > 1 %then
				%do;
					%let end = %eval(&varend -1);
					data section&i; set section&i;
						drop var&varend-var&nvar;
						%do j = 1 %to &end;
							%let k = %eval(&offset + &j);
							rename var&j = var&k;
							%let MLen&k = &&MLen&j;
						%end;
					run;
				%end;
			%else 
				%do;
					data section&i; set section&i;
						drop var&varend-var&nvar;						
					run;
				%end;
			%let offset = %eval(&offset + &varend - 1);			
		%end;
		data gdadata; merge popfreqstr &mergestring;
		run;
	%end;

%if &HasID %then 
	%do;
		data outdata; set gdadata;	
			if flag = 1 then delete;
			length Population $ &PopLen IndividualID $ &IDLen;	
			%do i = 1 %to &NLoci;
				length Mk&i $ &&MLen&i.; *&&Marker&i. $ &&MLen&i.;
			%end;
			Population = pop;
			IndividualID = var1; 
			%do i = 1 %to &NLoci;
				%let j = %eval(&i + 1);
				Mk&i = translate(var&j, &MissingAllele, "&MissingLabel");
			%end;		
			freq = freqstr + 0;	
			drop pop flag var1-var&nvar freqstr; 
		run;
	%end;
%else 
	%do;
		data outdata; set gdadata;	
			if flag = 1 then delete;
			length Population $ &PopLen;	
			%do i = 1 %to &NLoci;
				length Mk&i $ &&MLen&i.; *&&Marker&i. $ &&MLen&i.;
			%end;
			Population = pop;
			%do i = 1 %to &NLoci;
				Mk&i = translate(var&i, &MissingAllele, "&MissingLabel");
			%end;
			freq = freqstr + 0;	
			drop pop flag var1-var&nvar freqstr; 
		run;
	%end;
%if &IsHaploid %then 
	%do;
		data outdata; set outdata;
			%do i = 1 %to &NLoci;
				A&i = Mk&i;
			%end;
			drop Mk1-Mk&NLoci;
		run;
	%end;
%else 
	%do;
		data outdata; set outdata;
			%let N2 = %eval(2*&NLoci);
			%do i = 1 %to &NLoci;
				%let j = %eval(2*&i-1);
				%let k = %eval(2*&i);
				delim = index(Mk&i, "&Separator");
				A&j = substr(Mk&i, 1, delim - 1);
				A&k = substr(Mk&i, delim + 1);
			%end;
			drop delim Mk1-Mk&NLoci;
		run;
	%end;

%if &HasFreq = 0 %then
	%do;
		data outlib.&OutData; set outdata(drop = freq);
		run;
	%end;
%else
	%do;
		data outlib.&OutData; set outdata;
			do i = 1 to freq;	
				output;
			end;
			drop freq i;
		run;
	%end;

*Finishing;
%let day2 = %sysfunc(round(%sysevalf(%sysfunc(date()))));
%let VeryEnd = %sysfunc(round(%sysevalf(%sysfunc(time())),.01));
%if &day1 = &day2 %then %do;
   %let TotalTime = %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;
%else %do;
   %let TotalTime = %sysevalf((&day2-&day1)*86400) + %sysfunc(round(%sysevalf(&VeryEnd - &VeryStart),.01));
%end;  
data _null_;
   t = &TotalTime;
   put "The total importing time is" t time11.2;
run;
%exit:;

* Packaging;
%PackageBegin(&ProcessName);
%PackageInsertDataset(data from &ProcessName, OutLib, &OutData);
%if %upcase(&UploadRawFile)=YES %then %do;
  %PackageInsertFile(Raw Data,&TmpPath,&TmpName2,,,No);
%end;
%if %length(&MarkerData)>0 %then %do;
 %PackageInsertDataset(marker data from &ProcessName, OutLib, &MarkerData);
%end;
%PackagePublish(&OutPath,to_requester);
%PackageEnd;

%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %ClockStop(&ProcessName);
%mend ParseNexus;
%ParseNexus;
