/*
SAS Genetic Marker Input Engine

Name: Pedigree Input Engine
Author: Wendy Czika, SAS Institute Inc.
Support: Wendy Czika, SAS Institute Inc.
History:
         02Mar04 initial coding                          

Description: The Pedigree Input Engine imports 
pedigree data into a SAS data set, then uploads 
it to the warehouse.  Pedigree data in formats 
such as LINKAGE, QTDT, Genehunter, and FBAT are 
accommodated by this input engine.  
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

OPTIONS nospool;

*ProcessVariables;
%let InFile = ;         
* Pedigree file with full path;
%let Delimiter = SPACE; 
* Column delimiter: SPACE, COMMA, or TAB;
%let NamesfromFirstRow = No;     
* Get column names from first row of file?;
%let VarNameList = Pedigree Individual Father Mother Sex ; 
* If NamesfromFirstRow=NO, names of variables in order corresponding to order of columns in InFile
can be specified to replace the default names used by PROC IMPORT, VAR1-VARn.
Ranges of variables such as A1-A10 can be specified with NO spaces before or after the hyphen.;
%let QuantVars = ;      
* Variables that are quantitative.  These variable names come from the first row of the
pedigree file if GetNames=Yes or from the Variable Name List field if GetNames=No.  Any
non-numeric values in these variables are set to missing. All variables in the Variable Names field
that are NOT specified as quantitative variables are set to missing when 0.;
%let PhenoInFile = ;    
* Optional file containing additional phenotypes/covariates.  Variables in this file
other than the Merge Variables are treated as quantitative variables.;
%let PhenoVarNameList = Pedigree Individual ;  
* If GetNames=No, names of all variables in order corresponding to order of columns in PhenoInFile.
Ranges of variables such as A1-A10 can be specified with NO spaces before or after the hyphen.
If this field is left blank, variables are named VAR1-VARn.;
%let MergeVars = Pedigree Individual;
* If PhenoInFile is used, variables with which to merge together InFile and PhenoInFile;
%let OutData = ;
* Name of SAS data set created from pedigree file and uploaded to the warehouse.;
%let OutPath = ; 
* Path for output on your SAS server. If you leave this blank the system 
will use a default location.;
%let UploadRawFiles = No; 
* Specify "Yes" to upload raw file(s) (in original format)
into warehouse;
%let MacroPath = ; 
* Path for SAS macros.;

*ProcessBody;

%macro PedigreeIE;

*initialization;
%let ProcessName = PedigreeInputEngine;

%global exiterror;
%let exiterror = 0;
%let clockrunning = 0;
options nospool errors=max;
data _null_;
   infile "&MacroPath.UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for SAS macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath.UtilityMacros.sas" / nosource;
%ClockStart(&ProcessName);
%if &syserr %then %goto exit;
%let clockrunning = 1;
%CheckForBlank(&InFile,pedigree file);
%if &exiterror %then %goto exit;
%PathName(&InFile);
%if &exiterror %then %do;
   %put ERROR: Pedigree file specification is invalid.;
   %goto exit;
%end;

* define libnames;
libname OutLib "&OutPath";

%CheckDataName(_name=OutData,_default=&ProcessName);

%if (&Delimiter eq SPACE) %then %do;
 filename fileref1 "&InFile";
 filename fileref2 "&OutPath.&TmpName._compbl";
%end;
%else %do; 
  filename fileref2 "&InFile";
%end;
 
%if (&Delimiter eq SPACE) %then %do;
 data _null_;
  infile fileref1 truncover lrecl=32767;
  file fileref2;
  input a $32767.;
  a = compbl(a);
  if length(a)>1 then put a;
 run;
%end;
  
PROC IMPORT OUT=&OutData
     DATAFILE=fileref2
     DBMS=dlm REPLACE;
     %if (&Delimiter ne SPACE) %then %do;
      %if (&Delimiter eq COMMA) %then %do;
       Delimiter=',';
      %end;
      %else %do;
       Delimiter='090d'X;
      %end;
     %end;
     GETNAMES=&NamesfromFirstRow; 
RUN;
 
%if (%upcase(&NamesfromFirstRow)=NO and %length(&VarNameList)>0) %then %do;
 
 %NObsVars(&OutData);
 %let rs = ;
 %let wc=1;
 %do i=1 %to &nvars;
  %let new=%scan(&VarNameList,&wc,%str( ));
  %let wc=%eval(&wc+1);
  %if %length(&new)>0 %then %do;
   %let hyp=%index(&new,%str(-));
   %if &hyp>0 %then %do;
    %let fnumloc=%sysfunc(indexc(&new,'0123456789'));
    %let prefix=%substr(&new,1,%eval(&fnumloc-1));
    %let snumloc=%eval(%sysfunc(indexc(%substr(&new,%eval(&hyp+1)),'0123456789'))+&hyp);
    %let num1=%substr(&new,&fnumloc,%eval(&hyp-&fnumloc));
	%let num2=%substr(&new,&snumloc);
	%let rng=%eval(&num2-&num1);
	%let last=%eval(&i+&rng);
	%let rs=%str(&rs VAR&i-VAR&last=&prefix.&num1-&prefix.&num2);
	%let i=%eval(&i+&rng);
   %end;
   %else %do;
    %let rs=%str(&rs VAR&i=&new);
   %end;
  %end;
 %end;
   
 data &OutData;
  set &OutData;
  rename &rs;
 run; 
  
%end;
 
%if (%length(&PhenoInFile)>0) %then %do;
 %PathName(&PhenoInFile);
 %if &exiterror %then %do;
  %put ERROR: Phenotype file specification is invalid.;
  %goto exit;
 %end;
 %if (&Delimiter eq SPACE) %then %do;
  filename fileref1 "&PhenoInFile";
  filename fileref2 "&OutPath.&TmpName._compbl";
 %end;
 %else %do;
  filename fileref2 "&PhenoInFile";
 %end;

 %if (&Delimiter eq SPACE) %then %do;
  data _null_;
   infile fileref1 truncover lrecl=32767;
   file fileref2;
   input a $32767.;
   a = compbl(a);
   if length(a)>1 then put a;
  run;
 %end;
  
 PROC IMPORT OUT=PhenoData
      DATAFILE=fileref2
      DBMS=dlm REPLACE;
	  %if (&Delimiter ne SPACE) %then %do;
	   %if (&Delimiter eq COMMA) %then %do; 
        Delimiter=',';
       %end;
	   %else %do;
        Delimiter='090d'X;
       %end;
	  %end;
	  GETNAMES=&NamesfromFirstRow; 
 RUN;
 
 %if (%upcase(&NamesfromFirstRow)=NO and %length(&PhenoVarNameList)>0) %then %do;
 
  %NObsVars(PhenoData);
  %let rs = ;
  %let wc=1;
  %do i=1 %to &nvars;
   %let new=%scan(&PhenoVarNameList,&wc,%str( ));
   %let wc=%eval(&wc+1);
   %if %length(&new)>0 %then %do;
    %let hyp=%index(&new,%str(-));
    %if &hyp>0 %then %do;
     %let fnumloc=%sysfunc(indexc(&new,'0123456789'));
     %let prefix=%substr(&new,1,%eval(&fnumloc-1));
     %let snumloc=%eval(%sysfunc(indexc(%substr(&new,%eval(&hyp+1)),'0123456789'))+&hyp);
	 %let num1=%substr(&new,&fnumloc,%eval(&hyp-&fnumloc));
	 %let num2=%substr(&new,&snumloc);
	 %let rng=%eval(&num2-&num1);
	 %let last=%eval(&i+&rng);
	 %let rs=%str(&rs VAR&i-VAR&last=&prefix.&num1-&prefix.&num2);
	 %let i=%eval(&i+&rng);
    %end;
    %else %let rs=%str(&rs VAR&i=&new);
   %end;
  %end;
   
  data PhenoData;
   set PhenoData;
   rename &rs;
  run;   
 %end;

 data PhenoSub;
  set PhenoData;
  drop &MergeVars;
 run;
 
 %let cvct=0;
 %let rs= ;
 %let dsid=%sysfunc(open(PhenoSub));
 %let ncv=%sysfunc(attrn(&dsid,nvars));
 %do i=1 %to &ncv;
  %if (%sysfunc(vartype(&dsid,&i)) = C) %then %do;
   %let cvct=%eval(&cvct+1);  
   %let currvar=%sysfunc(varname(&dsid,&i));
   %let rs=%str(&rs newqvar&cvct=&currvar);
  %end;
 %end;   
 %let rc1=%sysfunc(close(&dsid));

 %if (&cvct>0) %then %do;
  data convertqvars;
   set PhenoSub;
   array qvars{&cvct} _CHARACTER_;
   array newqvar{&cvct};
   do i=1 to &cvct;
    newqvar{i}=qvars{i}+0;
   end;
   drop _CHARACTER_;
  run;

  data convertqvars;
   set convertqvars;
   rename &rs;
  run;
 %end;
 %else %do;
  data convertqvars;
   set PhenoSub;
  run;
 %end;
 
 data mrgvars;
  set PhenoData(keep=&MergeVars);
  array numv{*} _NUMERIC_;
  array numc{*} _CHARACTER_;
  do i=1 to dim(numv);
   if numv{i}=0 then numv{i}=.;
  end;
  do i=1 to dim(numc);
   if numc{i}='0' then numc{i}=' ';
  end;
 run;
 
 data otherqvars;
  merge mrgvars convertqvars;
 run;
 
 %if %length(&QuantVars)>0 %then %do;
  data qsub;
   set &OutData;
   keep &QuantVars;
  run;

  %let cvct=0;
  %let rs= ;
  %let dsid=%sysfunc(open(qsub));
  %let ncv=%sysfunc(attrn(&dsid,nvars));
  %do i=1 %to &ncv;
   %if (%sysfunc(vartype(&dsid,&i)) = C) %then %do;
    %let cvct=%eval(&cvct+1);
    %let currvar=%sysfunc(varname(&dsid,&i));
    %let rs=%str(&rs newqvar&cvct=&currvar);
   %end;
  %end;
  %let rc1=%sysfunc(close(&dsid));

  %if (&cvct>0) %then %do;
   data convertqvars;
    set qsub;
    array qvars{*} _CHARACTER_;
    array newqvar{&cvct};
    do i=1 to &cvct;
     newqvar{i}=qvars{i}+0;
    end;
    drop _CHARACTER_;
   run;

   data convertqvars;
    set convertqvars;
    rename &rs;
   run;
  %end;
  %else %do;
   data convertqvars;
    set qsub;
   run;
  %end;

  data convertcvars;
   set &OutData(drop=&QuantVars);
   array numv{*} _NUMERIC_;
   array numc{*} _CHARACTER_;
   do i=1 to dim(numv);
    if numv{i}=0 then numv{i}=.;
   end;
   do i=1 to dim(numc);
    if numc{i}='0' then numc{i}=' ';
   end;
  run; 
 
  data allinvars;
   merge convertcvars convertqvars;
  run;  
 %end;

 %else %do;
  data allinvars;
   set &OutData;
   array numv{*} _NUMERIC_;
   array numc{*} _CHARACTER_;
   do i=1 to dim(numv);
    if numv{i}=0 then numv{i}=.;
   end;
   do i=1 to dim(numc);
    if numc{i}='0' then numc{i}=' ';
   end;
  run; 
 %end;
 
 proc sort data=otherqvars;
  by &MergeVars;
 run;
 
 proc sort data=allinvars;
  by &MergeVars;
 run;
 
 data OutLib.&OutData;
  merge allinvars otherqvars;
  by &MergeVars;
  drop i;
 run;  
%end;

%else %do;
 %if %length(&QuantVars)>0 %then %do;
  data qsub;
   set &OutData;
   keep &QuantVars;
  run;

  %let rs= ;
  %let cvct=0;
  %let dsid=%sysfunc(open(qsub));
  %let ncv=%sysfunc(attrn(&dsid,nvars));
  %do i=1 %to &ncv;
   %if (%sysfunc(vartype(&dsid,&i)) = C) %then %do;
    %let cvct=%eval(&cvct+1);
    %let currvar=%sysfunc(varname(&dsid,&i));
    %let rs=%str(&rs newqvar&cvct=&currvar);
   %end;
  %end;   
  %let rc1=%sysfunc(close(&dsid));

  %if (&cvct>0) %then %do;
   data convertqvars;
    set qsub;
    array qvars{*} _CHARACTER_;
    array newqvar{&cvct};
    do i=1 to &cvct;
     newqvar{i}=qvars{i}+0;
    end;
    drop _CHARACTER_;
   run;

   data convertqvars;
    set convertqvars;
    rename &rs;
   run;
  %end;
  %else %do;
   data convertqvars;
    set qsub;
   run;
  %end;

  data convertcvars;
   set &OutData(drop=&QuantVars);
   array numv{*} _NUMERIC_;
   array numc{*} _CHARACTER_;
   do i=1 to dim(numv);
    if numv{i}=0 then numv{i}=.;
   end;
   do i=1 to dim(numc);
    if numc{i}='0' then numc{i}=' ';
   end;
  run; 
 
  data OutLib.&OutData;
   merge convertcvars convertqvars;
   drop i;
  run;
 %end;
 %else %do;
   data OutLib.&OutData;
    set &OutData;
    array numv{*} _NUMERIC_;
    array numc{*} _CHARACTER_;
    do i=1 to dim(numv);
     if numv{i}=0 then numv{i}=.;
    end;
    do i=1 to dim(numc);
     if numc{i}='0' then numc{i}=' ';
    end;
	drop i;
   run; 
 %end;
%end;
   
%NObsVars(OutLib.&OutData);
%if (&nobs=0 | &nvars=0) %then %do;
   data OutLib.exitedwitherrors;
      error=1;
   run;
   %put ERROR: Error(s) were encountered creating &OutData data set.;
   %let exiterror = 1;
   %goto exit;
%end;

* Packaging;
%PackageBegin(&ProcessName);
%PackageInsertDataset(data from &ProcessName, OutLib, &OutData);
%if %upcase(&UploadRawFiles)=YES %then %do;
   %PathName(&InFile);
   %PackageInsertFile(raw pedigree data,&TmpPath,&TmpName2,,,No);
   %if %length(&PhenoInFile) %then %do;
    %PathName(&PhenoInFile);
    %PackageInsertFile(raw phenotype data,&TmpPath,&TmpName2,,,No);
   %end;
%end;
%PackagePublish(&OutPath,to_requester);
%PackageEnd;

%exit:;
%if (&syserr | &exiterror) %then %put ERROR: &ProcessName exited due to errors.;
%if (&clockrunning) %then %ClockStop(&ProcessName);

%mend PedigreeIE;
%PedigreeIE;







